#include <sqlextract/filter/filter_core.h>
#include <sqlextract/filter/filter_arithmetic.h>
#include <sqlextract/filter/arbval.h>
#include <sqlextract/util.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

struct filter_arithmetic_storage_t {
	struct filter_t *lhs;
	struct filter_t *rhs;
};

static void filter_arithmetic_destroy(void *storage) {
	struct filter_arithmetic_storage_t *stg =
			(struct filter_arithmetic_storage_t *) storage;

	filter_destroy(stg->lhs);
	filter_destroy(stg->rhs);
	free(stg);
}

/* This method needs some explanation...
 *
 * All operators (at least +, -, * and /) uses the same type structure, which is
 * return integer unless one is real, then return real.
 *
 * The concept is therefore, given two filters, with objects, return the lhs,
 * rhs value, and set the variable type of the return value to all the same,
 * and return it's type.
 *
 * All methods uses this logic in the beginning anyway.
 */
static enum arbval_type_t filter_arithmetic_preprocess(struct arbval_t **result,
		struct arbval_t **lhsval, struct arbval_t **rhsval, void *object,
		void *storage) {
	struct filter_arithmetic_storage_t *stg =
			(struct filter_arithmetic_storage_t *) storage;

	*lhsval = filter_process(stg->lhs, object);
	*rhsval = filter_process(stg->rhs, object);

	if ((*lhsval)->type == ARBVAL_TYPE_REAL
			|| (*rhsval)->type == ARBVAL_TYPE_REAL) {
		arbval_typecast_real(*lhsval);
		arbval_typecast_real(*rhsval);
		arbval_typecast_real(*result);
	} else {
		arbval_typecast_integer(*lhsval);
		arbval_typecast_integer(*rhsval);
		arbval_typecast_integer(*result);
	}
	return (*result)->type;
}

static void filter_process_add(struct arbval_t *result, void *object,
		void *storage) {
	struct arbval_t *lhsval;
	struct arbval_t *rhsval;
	switch (filter_arithmetic_preprocess(&result, &lhsval, &rhsval, object,
			storage)) {
	case ARBVAL_TYPE_NULL:
	case ARBVAL_TYPE_STRING:
		break;
	case ARBVAL_TYPE_REAL:
		result->value.real = lhsval->value.real + rhsval->value.real;
		break;
	case ARBVAL_TYPE_INTEGER:
		result->value.integer = lhsval->value.integer + rhsval->value.integer;
		break;
	}
}

struct filter_t *filter_create_add(struct filter_t *lhs, struct filter_t *rhs) {
	struct filter_arithmetic_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_arithmetic_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_add, filter_arithmetic_destroy,
			(void*) stg);
}

static void filter_process_sub(struct arbval_t *result, void *object,
		void *storage) {
	struct arbval_t *lhsval;
	struct arbval_t *rhsval;
	switch (filter_arithmetic_preprocess(&result, &lhsval, &rhsval, object,
			storage)) {
	case ARBVAL_TYPE_NULL:
	case ARBVAL_TYPE_STRING:
		break;
	case ARBVAL_TYPE_REAL:
		result->value.real = lhsval->value.real - rhsval->value.real;
		break;
	case ARBVAL_TYPE_INTEGER:
		result->value.integer = lhsval->value.integer - rhsval->value.integer;
		break;
	}
}

struct filter_t *filter_create_sub(struct filter_t *lhs, struct filter_t *rhs) {
	struct filter_arithmetic_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_arithmetic_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_sub, filter_arithmetic_destroy,
			(void*) stg);
}

static void filter_process_mul(struct arbval_t *result, void *object,
		void *storage) {
	struct arbval_t *lhsval;
	struct arbval_t *rhsval;
	switch (filter_arithmetic_preprocess(&result, &lhsval, &rhsval, object,
			storage)) {
	case ARBVAL_TYPE_NULL:
	case ARBVAL_TYPE_STRING:
		break;
	case ARBVAL_TYPE_REAL:
		result->value.real = lhsval->value.real * rhsval->value.real;
		break;
	case ARBVAL_TYPE_INTEGER:
		result->value.integer = lhsval->value.integer * rhsval->value.integer;
		break;
	}
}

struct filter_t *filter_create_mul(struct filter_t *lhs, struct filter_t *rhs) {
	struct filter_arithmetic_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_arithmetic_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_mul, filter_arithmetic_destroy,
			(void*) stg);
}

static void filter_process_div(struct arbval_t *result, void *object,
		void *storage) {
	struct arbval_t *lhsval;
	struct arbval_t *rhsval;
	switch (filter_arithmetic_preprocess(&result, &lhsval, &rhsval, object,
			storage)) {
	case ARBVAL_TYPE_NULL:
	case ARBVAL_TYPE_STRING:
		break;
	case ARBVAL_TYPE_REAL:
		if (rhsval->value.real == 0.0) {
			/* We don't have any means test for exceptions, and we don't really
			 * intend to either. Just define it as the hugest possible value
			 */
			result->value.real = INFINITY;
		} else {
			result->value.real = lhsval->value.real / rhsval->value.real;
		}
		break;
	case ARBVAL_TYPE_INTEGER:
		if (rhsval->value.integer == 0) {
			/* We don't have any means test for exceptions, and we don't really
			 * intend to either. Just define it as the hugest possible value
			 */
			result->value.integer = LONG_MAX;
		} else {
			result->value.integer = lhsval->value.integer / rhsval->value.integer;
		}
		break;
	}
}

struct filter_t *filter_create_div(struct filter_t *lhs, struct filter_t *rhs) {
	struct filter_arithmetic_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_arithmetic_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_div, filter_arithmetic_destroy,
			(void*) stg);
}

static void filter_process_mod(struct arbval_t *result, void *object,
		void *storage) {
	struct arbval_t *lhsval;
	struct arbval_t *rhsval;
	switch (filter_arithmetic_preprocess(&result, &lhsval, &rhsval, object,
			storage)) {
	case ARBVAL_TYPE_NULL:
	case ARBVAL_TYPE_STRING:
		break;
	case ARBVAL_TYPE_REAL:
		if (rhsval->value.real == 0.0) {
			/* We don't have any means test for exceptions, and we don't really
			 * intend to either. Just define it as the hugest possible value
			 */
			result->value.real = INFINITY;
		} else {
			/* sqlite processes it in an int domain... */
			result->value.real = (long)(lhsval->value.real) % (long)(rhsval->value.real);
		}
		break;
	case ARBVAL_TYPE_INTEGER:
		if (rhsval->value.integer == 0) {
			/* We don't have any means test for exceptions, and we don't really
			 * intend to either. Just define it as the hugest possible value
			 */
			result->value.integer = LONG_MAX;
		} else {
			result->value.integer = lhsval->value.integer % rhsval->value.integer;
		}
		break;
	}
}

struct filter_t *filter_create_mod(struct filter_t *lhs, struct filter_t *rhs) {
	struct filter_arithmetic_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_arithmetic_storage_t));
	stg->lhs = lhs;
	stg->rhs = rhs;
	return filter_create(filter_process_mod, filter_arithmetic_destroy,
			(void*) stg);
}

static void filter_process_neg(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_t *subf = (struct filter_t *) storage;
	struct arbval_t *subval;

	subval = filter_process(subf, object);
	switch (subval->type) {
	case ARBVAL_TYPE_REAL:
		arbval_typecast_real(result);
		result->value.real = -subval->value.real;
		break;
	case ARBVAL_TYPE_STRING:
		arbval_typecast_integer(subval);
		/* no break */
	case ARBVAL_TYPE_INTEGER:
		arbval_typecast_integer(result);
		result->value.integer = -subval->value.integer;
		break;
	case ARBVAL_TYPE_NULL:
		arbval_typecast_integer(result);
		result->value.integer = 0;
		break;
	}
}

static void filter_destroy_neg(void *storage) {
	struct filter_t *subf = (struct filter_t *) storage;
	filter_destroy(subf);
}
struct filter_t *filter_create_neg(struct filter_t *subfilter) {
	return filter_create(filter_process_neg, filter_destroy_neg,
			(struct filter_t *) subfilter);
}
